/*
 * new_comm.c
 *
 *  Created on: Jan 2, 2017
 *      Author: dawehr
 */
#include "new_comm.h"

#define INTC		XScuGic
#define COMM_UART_DEVICE_ID		XPAR_PS7_UART_0_DEVICE_ID
#define COMM_INTC_DEVICE_ID		XPAR_SCUGIC_SINGLE_DEVICE_ID
#define COMM_UART_INT_IRQ_ID		XPAR_PS7_UART_0_INTR//XPAR_XUARTPS_0_INTR

#define BAUD_RATE 115200
// Maximum number of bytes to be received within our loop time, plus 64 for a little buffer
// (bit/s) * (seconds) / (10 bits / byte for UART)
#define UART_BUF_SIZE (((BAUD_RATE * (DESIRED_USEC_PER_LOOP / 1000) / 1000) / 10) + 64)
#define MAX_PACKET_SIZE 256

// Declaration of interrupt handler
void Handler(void *CallBackRef, u32 Event, unsigned int EventData);

// Pointer to the UART driver instance
static XUartPs* uartInstPtr;

static u8 recvBuf[UART_BUF_SIZE];
static volatile size_t received_in_buf = 0;

static u8 secondaryBuf[UART_BUF_SIZE + MAX_PACKET_SIZE];
static volatile size_t secondaryEnd = 0; // Index of start of valid data in the secondary buffer
static size_t secondaryBegin = 0; // Index of one past end of valid data in the secondary buffer

int initUartComms() {
	uartInstPtr = uart0_init(COMM_UART_DEVICE_ID, BAUD_RATE);
	// Initialize UART0 (Bluetooth/WiFi)
	if(!uartInstPtr) {
		return -1;
	}

	uart0_clearFIFOs();

	if (uart0_int_init(COMM_UART_INT_IRQ_ID, (Xil_ExceptionHandler) uartInterruptHandler) != XST_SUCCESS) {
		return -1;
	}


	/*
	 * Setup the handlers for the UART that will be called from the
	 * interrupt context when data has been sent and received, specify
	 * a pointer to the UART driver instance as the callback reference
	 * so the handlers are able to access the instance data
	 */
	//XUartPs_SetHandler(uartInstPtr, (XUartPs_Handler)Handler, uartInstPtr);

	u32 IntrMask = XUARTPS_IXR_RXFULL | XUARTPS_IXR_RXOVR | XUARTPS_IXR_TOUT;

	XUartPs_SetInterruptMask(uartInstPtr, IntrMask);


	/*
	 * Set the receiver timeout. If it is noDidn'tt set, and the last few bytes
	 * of data do not trigger the over-water or full interrupt, the bytes
	 * will not be received. By default it is disabled.
	 * Timeout duration = RecvTimeout x 4 x Bit Period. 0 disables the
	 *		timeout function.
	 *
	 * The setting of 8 will timeout after 8 x 4 = 32 character times.
	 * Increase the time out value if baud rate is high, decrease it if
	 * baud rate is low.
	 */
	XUartPs_SetRecvTimeout(uartInstPtr, 4);

	// Second argument is the number of bytes to trigger an interrupt at
	XUartPs_SetFifoThreshold(uartInstPtr, 12);

	//XUartPs_Recv(uartInstPtr, recvBuf, UART_BUF_SIZE);

	return 0;
}
// TODO: Be consistent with underscore_names or camelCase

int process_packet(unsigned char* packet, modular_structs_t *structs) {
	metadata_t meta_data;
	// parse metadata
	meta_data.begin_char = packet[0];
	meta_data.msg_type = packet[1];
	meta_data.msg_subtype = packet[2];
	meta_data.msg_id = (packet[4] << 8) | (packet[3]);
	meta_data.data_len = (packet[6] << 8) | (packet[5]);
	unsigned char packet_checksum = packet[7+meta_data.data_len];
	unsigned char* packet_data = packet + sizeof(metadata_t);

	// Compute checksum
	int i;
	unsigned char calculated_checksum = 0;
	for(i = 0; i < meta_data.data_len + 7; i++){
		calculated_checksum ^= packet[i];
	}
	// Discard if checksum didn't match
	if(packet_checksum != calculated_checksum) {
		printf("Checksums did not match: 0x%02x\t0x%02x\n", packet_checksum, calculated_checksum);
		return -1;
	}

	// Call appropriate function for packet
	(* (MessageTypes[meta_data.msg_type].subtypes[meta_data.msg_subtype].functionPtr))(packet_data, meta_data.data_len, structs);

	return 0;
}

void parse_available(modular_structs_t *structs) {
	// Discard all data before packet begin character
	while (secondaryBegin < secondaryEnd && secondaryBuf[secondaryBegin] != BEGIN_CHAR) {
		secondaryBegin++;
	}

	// Minimum size of a packet (header + checksum)
	int min_packet_size = sizeof(metadata_t) + 1;
	// TODO: Loop limit?
	while (secondaryEnd - secondaryBegin >= min_packet_size) {
		unsigned char* packet_header = secondaryBuf + secondaryBegin;
		size_t packet_len = (packet_header[6] << 8) | (packet_header[5]);
		if (secondaryEnd - secondaryBegin >= min_packet_size + packet_len) {
			process_packet(secondaryBuf + secondaryBegin, structs);
			secondaryBegin += min_packet_size + packet_len;
		}
		else {
			// Not enough data for a packet
			break;
		}
	}
}

u32 disable_interrupts() {
	u32 ImrRegister;
	ImrRegister = XUartPs_ReadReg(uartInstPtr->Config.BaseAddress, XUARTPS_IMR_OFFSET);
	XUartPs_WriteReg(uartInstPtr->Config.BaseAddress, XUARTPS_IDR_OFFSET, XUARTPS_IXR_MASK);
	return ImrRegister;
}

void restore_interrupts(u32 intr_state) {
	XUartPs_WriteReg(uartInstPtr->Config.BaseAddress, XUARTPS_IER_OFFSET, intr_state);
}

void process_received(modular_structs_t *structs) {
//	// Move unprocessed bytes to front of secondary buffer
//	size_t unprocessed_size = secondaryEnd - secondaryBegin;
//	memmove(secondaryBuf, secondaryBuf + secondaryBegin, unprocessed_size);
//	secondaryBegin = 0;
//	secondaryEnd = unprocessed_size;
//
//	// Temporarily disable the receive interrupt
//	XUartPs_Recv(uartInstPtr, recvBuf, 0);
//	if (received_in_buf) {
//		volatile int a = 3;
//	}
//	// Put received data into secondary buffer
//	memcpy(secondaryBuf + secondaryEnd, recvBuf, received_in_buf);
//
//	secondaryEnd += received_in_buf;
//	received_in_buf = 0;
//	// Re-enable interrupts and process copied data
//	XUartPs_Recv(uartInstPtr, recvBuf, UART_BUF_SIZE);
	if (secondaryEnd != 0) {
		volatile int asdasdf = 3;
		MIO7_led_off();
	}

	parse_available(structs);

	u32 intr_state = disable_interrupts();

	// Move unprocessed bytes to front of secondary buffer
	size_t unprocessed_size = secondaryEnd - secondaryBegin;
	memmove(secondaryBuf, secondaryBuf + secondaryBegin, unprocessed_size);
	secondaryBegin = 0;
	secondaryEnd = unprocessed_size;

	restore_interrupts(intr_state);
	//unsigned char in_fifo = XUartPs_ReadReg(uartInstPtr->Config.BaseAddress, XUARTPS_FIFO_OFFSET);
	return;
}

void uartInterruptHandler(XUartPs *InstancePtr) {
	u32 IsrStatus;

	/*
	 * Read the interrupt ID register to determine which
	 * interrupt is active
	 */
	IsrStatus = XUartPs_ReadReg(InstancePtr->Config.BaseAddress,
				   XUARTPS_IMR_OFFSET);

	IsrStatus &= XUartPs_ReadReg(InstancePtr->Config.BaseAddress,
				   XUARTPS_ISR_OFFSET);

	u32 CsrRegister;
	/*
	* Read the Channel Status Register to determine if there is any data in
	 * the RX FIFO
	 */
	CsrRegister = XUartPs_ReadReg(InstancePtr->Config.BaseAddress,
				XUARTPS_SR_OFFSET);

	while (0 == (CsrRegister & XUARTPS_SR_RXEMPTY)) {
		secondaryBuf[secondaryEnd] = XUartPs_ReadReg(InstancePtr->Config.BaseAddress, XUARTPS_FIFO_OFFSET);
		secondaryEnd += 1;
		CsrRegister = XUartPs_ReadReg(InstancePtr->Config.BaseAddress, XUARTPS_SR_OFFSET);
		MIO7_led_on();
	}

	// Clear the interrupt status.
	XUartPs_WriteReg(InstancePtr->Config.BaseAddress, XUARTPS_ISR_OFFSET,
			IsrStatus);
}

/**************************************************************************/
/**
*
* This function is the handler which performs processing to handle data events
* from the device.  It is called from an interrupt context. so the amount of
* processing should be minimal.
*
* This handler provides an example of how to handle data for the device and
* is application specific.
*
* @param	CallBackRef contains a callback reference from the driver,
*		in this case it is the instance pointer for the XUartPs driver.
* @param	Event contains the specific kind of event that has occurred.
* @param	EventData contains the number of bytes sent or received for sent
*		and receive events.
*
* @return	None.
*
* @note		None.
*
***************************************************************************/
void Handler(void *CallBackRef, u32 Event, unsigned int EventData)
{
	/* data has been received */
	if (Event == XUARTPS_EVENT_RECV_DATA || Event == XUARTPS_EVENT_RECV_TOUT
			|| Event == XUARTPS_IXR_RXOVR || Event == XUARTPS_IXR_RXFULL) {
		received_in_buf = EventData;
	}
}

